/*
 * rdate.c: rdate for Win32
 * 
 * As ever, programming in Win32 is a bit of a struggle and the
 * result is not typically worth it.
 * 
 * Author: Chris Lightfoot <chris@ex-parrot.com>
 *         http://www.ex-parrot.com/~chris/
 *         
 * Donated into the public domain, 2000.
 * 
 */
#include <winsock.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#define PROGNAME	"rdate"

#define EPOCH_OFFSET	2208988800 /* RFC 868 */

/* Windows Sockets is TOO DUMB to set errno, so we can't readily produce useful diagnostics. */
#define perror(foo)		fprintf(stderr, foo": there was a Windows Sockets error of type %d\n", WSAGetLastError())

void printtime(struct tm *tim) {
	char *days[] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
	char *months[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
	
	printf("%s %s %02d %02d:%02d:%02d %04d GMT\n", days[tim->tm_wday], months[tim->tm_mon], tim->tm_mday, tim->tm_hour, tim->tm_min, tim->tm_sec, tim->tm_year + 1900);
}

int main(int argc, char **argv) {
	SOCKET s;
	WSADATA wsa;
	char **a;
	char *hostname = NULL;
	int setclock = 0;
	HOSTENT *he;
	SOCKADDR_IN sin;
	unsigned long T;
	char *x;
	struct tm *remotetime;

	memset(&sin, 0, sizeof(sin));

	sin.sin_family = AF_INET;
	sin.sin_port = htons(37);	/* RFC868 time service */

	if (WSAStartup(MAKEWORD(1, 1), &wsa)) {
		perror(PROGNAME": WSAStartup");
		return 1;
	}

	for (a = argv + 1; a < argv + argc; a++) {
		if (!strcmp(*a, "-s")) setclock = 1;
		else if (!strcmp(*a, "-h")) {
			printf(PROGNAME": find out the time on a remote system via RFC868 time service\n"
				   "\n"
				   "    "PROGNAME" [-h] [-s] <hostname>\n"
				   "        -h    print this message\n"
				   "        -s    set local clock\n"
				   "\n"
				   "Author: Chris Lightfoot <chris@ex-parrot.com>, http://www.ex-parrot.com/~chris/\n"
				   "Donated into the public domain, 2000\n");
			return 0;
		} else if (!hostname)
			hostname = *a;
		else {
			fprintf(stderr, "Invalid option \"%s\"\n", *a);
			return 1;
		}
	}

	if (!hostname) {
		fprintf(stderr, "No hostname supplied\n");
		return 1;
	}

	/* Try host as numeric, then symbolic. */
	sin.sin_addr.S_un.S_addr = inet_addr(hostname);

	if (sin.sin_addr.S_un.S_addr == INADDR_NONE) {
		he = gethostbyname(hostname);

		if (!he) {
			fprintf(stderr, "Host \"%s\" is unknown\n", hostname);
			return 1;
		}

		memcpy(&(sin.sin_addr), he->h_addr, sizeof(sin.sin_addr));
	}

	if ((s = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET) {
		perror(PROGNAME": socket");
		return 1;
	}

	if (connect(s, (SOCKADDR*)&sin, sizeof(sin)) == SOCKET_ERROR) {
		perror(PROGNAME": connect");
		return 1;
	}

	for (x = (char*)&T; x < (char*)&T + sizeof(T); x += recv(s, x, x - (char*)&T + sizeof(T), 0));

	closesocket(s);

	T = ntohl(T);

	T -= EPOCH_OFFSET;

	/* This is nasty and might not work. Need to correct for timezones because Windows is TOO DUMB
	 * to do the right thing and use GMT for its clock. We use libc's localtime(3) and hope for
	 * the best....
	 */
	remotetime = localtime(&T);

	if (setclock) {
		SYSTEMTIME st;

		st.wYear = remotetime->tm_year + 1900;
		st.wMonth = remotetime->tm_mon + 1;
		st.wDay = remotetime->tm_mday;
		st.wHour = remotetime->tm_hour;
		st.wMinute = remotetime->tm_min;
		st.wSecond = remotetime->tm_sec;
		st.wMilliseconds = 0;

		if (SetSystemTime(&st) == 0) {
			fprintf(stderr, PROGNAME": SetSytemTime: an error of type %x occured\n        Are you sure you have sufficient priveleges to set the system clock?\n", GetLastError());
			return 0;
		}
	}
	
	printtime(remotetime);

	return 0;
}
